#include <cassert>
#include "AVI.h"

//-----------------------------------------------------------------------------
//
//	RIFF header
//
//-----------------------------------------------------------------------------
RIFFHeader::RIFFHeader( unsigned long _fileSize )
{
	fourCC[ 0 ] = 'R';
	fourCC[ 1 ] = 'I';
	fourCC[ 2 ] = 'F';
	fourCC[ 3 ] = 'F';

	fileSize = _fileSize;

	fileType[ 0 ] = 'A';
	fileType[ 1 ] = 'V';
	fileType[ 2 ] = 'I';
	fileType[ 3 ] = ' ';
}

std::ostream& operator<<( std::ostream& _out, const RIFFHeader& _h )
{
	_out.write( reinterpret_cast< const char* >( _h.fourCC ), sizeof( _h.fourCC ) );
	_out.write( reinterpret_cast< const char* >( &_h.fileSize ), sizeof( _h.fileSize ) );
	_out.write( reinterpret_cast< const char* >( _h.fileType ), sizeof( _h.fileType ) );
	return _out;
}


ChunkHeader::ChunkHeader( const std::string& _fourCC, unsigned long _chunkSize )
{
	assert( _fourCC.length() == 4 );
	for( size_t i = 0; i < 4; ++i )
	{
		fourCC[ i ] = _fourCC[ i ];
	}

	chunkSize = _chunkSize;
}

std::ostream& operator<<( std::ostream& _out, const ChunkHeader& _h )
{
	_out.write( reinterpret_cast< const char* >( _h.fourCC ), sizeof( _h.fourCC ) );
	_out.write( reinterpret_cast< const char* >( &_h.chunkSize ), sizeof( _h.chunkSize ) );
	return _out;
}

ListHeader::ListHeader( const std::string& _listType, unsigned long _listSize )
{
	fourCC[ 0 ] = 'L';
	fourCC[ 1 ] = 'I';
	fourCC[ 2 ] = 'S';
	fourCC[ 3 ] = 'T';

	listSize = _listSize;

	assert( _listType.length() == 4 );
	for( size_t i = 0; i < 4; ++i )
	{
		listType[ i ] = _listType[ i ];
	}
}

std::ostream& operator<<( std::ostream& _out, const ListHeader& _h )
{
	_out.write( reinterpret_cast< const char* >( _h.fourCC ), sizeof( _h.fourCC ) );
	_out.write( reinterpret_cast< const char* >( &_h.listSize ), sizeof( _h.listSize ) );
	_out.write( reinterpret_cast< const char* >( _h.listType ), sizeof( _h.listType ) );
	return _out;
}

AVIMainHeader::AVIMainHeader( unsigned long _width, unsigned long _height, int _framesPerSecond, unsigned long _frameCount )
{
	fourCC[ 0 ] = 'a';
	fourCC[ 1 ] = 'v';
	fourCC[ 2 ] = 'i';
	fourCC[ 3 ] = 'h';

	int frameSize = _width * _height * 3;

	headerSize = 56;
	microSecondsPerFrame = static_cast< unsigned long >( 1000000.0 / _framesPerSecond );
	maxBytesPerSecond = frameSize * _framesPerSecond;
	paddingGranularity = 0;
	flags = 0x00000010;				// has index
	totalFrames = _frameCount;
	initialFrames = 0;
	streams = 1;
	suggestedBufferSize = 0;
	width = _width;
	height = _height;

	reserved[ 0 ] = 0;
	reserved[ 1 ] = 0;
	reserved[ 2 ] = 0;
	reserved[ 3 ] = 0;
}

std::ostream& operator<<( std::ostream& _out, const AVIMainHeader& _h )
{
	_out.write( reinterpret_cast< const char* >( _h.fourCC ), sizeof( _h.fourCC ) );
	_out.write( reinterpret_cast< const char* >( &_h.headerSize ), sizeof( _h.headerSize ) );
	_out.write( reinterpret_cast< const char* >( &_h.microSecondsPerFrame ), sizeof( _h.microSecondsPerFrame ) );
	_out.write( reinterpret_cast< const char* >( &_h.maxBytesPerSecond ), sizeof( _h.maxBytesPerSecond ) );
	_out.write( reinterpret_cast< const char* >( &_h.paddingGranularity ), sizeof( _h.paddingGranularity ) );
	_out.write( reinterpret_cast< const char* >( &_h.flags ), sizeof( _h.flags ) );
	_out.write( reinterpret_cast< const char* >( &_h.totalFrames ), sizeof( _h.totalFrames ) );
	_out.write( reinterpret_cast< const char* >( &_h.initialFrames ), sizeof( _h.initialFrames ) );
	_out.write( reinterpret_cast< const char* >( &_h.streams ), sizeof( _h.streams ) );
	_out.write( reinterpret_cast< const char* >( &_h.suggestedBufferSize ), sizeof( _h.suggestedBufferSize ) );
	_out.write( reinterpret_cast< const char* >( &_h.width ), sizeof( _h.width ) );
	_out.write( reinterpret_cast< const char* >( &_h.height ), sizeof( _h.height ) );
	_out.write( reinterpret_cast< const char* >( _h.reserved ), sizeof( _h.reserved ) );
	return _out;
}

AVIStreamHeader::AVIStreamHeader( int _width, int _height, int _framesPerSecond, int _frameCount )
{
	int padSize = 4 - ( ( _width * 3 ) % 4 );
	if( padSize == 4 ) padSize = 0;

	fourCC[ 0 ] = 's';
	fourCC[ 1 ] = 't';
	fourCC[ 2 ] = 'r';
	fourCC[ 3 ] = 'h';

	headerSize = 56;
	streamType[ 0 ] = 'v';
	streamType[ 1 ] = 'i';
	streamType[ 2 ] = 'd';
	streamType[ 3 ] = 's';

	streamHandler[ 0 ] = 'D';
	streamHandler[ 1 ] = 'I';
	streamHandler[ 2 ] = 'B';
	streamHandler[ 3 ] = ' ';

	flags = 0;
	priority = 0;
	language = 0;
	initialFrames = 0;
	scale = 1;
	rate = _framesPerSecond;
	start = 0;
	length = _frameCount;
	suggestedBufferSize = 8 + ( ( _width * 3 ) + padSize ) * _height;
	quality = 0;
	sampleSize = 0;

	frame.left = 0;
	frame.top = 0;
	frame.right = _width;
	frame.bottom = _height;
}

std::ostream& operator<<( std::ostream& _out, const AVIStreamHeader& _h )
{
	_out.write( reinterpret_cast< const char* >( _h.fourCC ), sizeof( _h.fourCC ) );
	_out.write( reinterpret_cast< const char* >( &_h.headerSize ), sizeof( _h.headerSize ) );
	_out.write( reinterpret_cast< const char* >( _h.streamType ), sizeof( _h.streamType ) );
	_out.write( reinterpret_cast< const char* >( _h.streamHandler ), sizeof( _h.streamHandler ) );
	_out.write( reinterpret_cast< const char* >( &_h.flags ), sizeof( _h.flags ) );
	_out.write( reinterpret_cast< const char* >( &_h.priority ), sizeof( _h.priority ) );
	_out.write( reinterpret_cast< const char* >( &_h.language ), sizeof( _h.language ) );
	_out.write( reinterpret_cast< const char* >( &_h.initialFrames ), sizeof( _h.initialFrames ) );
	_out.write( reinterpret_cast< const char* >( &_h.scale ), sizeof( _h.scale ) );
	_out.write( reinterpret_cast< const char* >( &_h.rate ), sizeof( _h.rate ) );
	_out.write( reinterpret_cast< const char* >( &_h.start ), sizeof( _h.start ) );
	_out.write( reinterpret_cast< const char* >( &_h.length ), sizeof( _h.length ) );
	_out.write( reinterpret_cast< const char* >( &_h.suggestedBufferSize ), sizeof( _h.suggestedBufferSize ) );
	_out.write( reinterpret_cast< const char* >( &_h.quality ), sizeof( _h.quality ) );
	_out.write( reinterpret_cast< const char* >( &_h.sampleSize ), sizeof( _h.sampleSize ) );
	_out.write( reinterpret_cast< const char* >( &_h.frame.left ), sizeof( _h.frame.left ) );
	_out.write( reinterpret_cast< const char* >( &_h.frame.top ), sizeof( _h.frame.top ) );
	_out.write( reinterpret_cast< const char* >( &_h.frame.right ), sizeof( _h.frame.right ) );
	_out.write( reinterpret_cast< const char* >( &_h.frame.bottom ), sizeof( _h.frame.bottom ) );
	return _out;
}

AVIStreamFormatHeader::AVIStreamFormatHeader( int _width, int _height )
{
	int padSize = 4 - ( ( _width * 3 ) % 4 );
	if( padSize == 4 ) padSize = 0;

	headerSize		= 40;
	width			= _width;
	height			= _height;
	colorPlanes		= 1;
	bitsPerPixel	= 24;
	compression		= 0;	// BI_RGB
	imageSize		= ( ( _width * 3 ) + padSize ) * _height;
	horizontalRes	= 0;
	verticalRes		= 0;
	colorCount		= 0;
	importantColors	= 0;
}

std::ostream& operator<<( std::ostream& _out, const AVIStreamFormatHeader& _h )
{
	_out.write( reinterpret_cast< const char* >( &_h.headerSize ), sizeof( _h.headerSize ) );
	_out.write( reinterpret_cast< const char* >( &_h.width ), sizeof( _h.width ) );
	_out.write( reinterpret_cast< const char* >( &_h.height ), sizeof( _h.height ) );
	_out.write( reinterpret_cast< const char* >( &_h.colorPlanes ), sizeof( _h.colorPlanes ) );
	_out.write( reinterpret_cast< const char* >( &_h.bitsPerPixel ), sizeof( _h.bitsPerPixel ) );
	_out.write( reinterpret_cast< const char* >( &_h.compression ), sizeof( _h.compression ) );
	_out.write( reinterpret_cast< const char* >( &_h.imageSize ), sizeof( _h.imageSize ) );
	_out.write( reinterpret_cast< const char* >( &_h.horizontalRes ), sizeof( _h.horizontalRes ) );
	_out.write( reinterpret_cast< const char* >( &_h.verticalRes ), sizeof( _h.verticalRes ) );
	_out.write( reinterpret_cast< const char* >( &_h.colorCount ), sizeof( _h.colorCount ) );
	_out.write( reinterpret_cast< const char* >( &_h.importantColors ), sizeof( _h.importantColors ) );
	return _out;
}

AVIIndexEntry::AVIIndexEntry( const std::string& _chunkID, int _chunkOffset, int _chunkSize )
{
	assert( _chunkID.length() == 4 );
	for( size_t i = 0; i < 4; ++i )
	{
		chunkID[ i ] = _chunkID[ i ];
	}

	flags = 0x00000010;		// keyframe
	offset = _chunkOffset;
	size = _chunkSize;
};

std::ostream& operator<<( std::ostream& _out, const AVIIndexEntry& _e )
{
	_out.write( reinterpret_cast< const char* >( _e.chunkID ), sizeof( _e.chunkID ) );
	_out.write( reinterpret_cast< const char* >( &_e.flags ), sizeof( _e.flags ) );
	_out.write( reinterpret_cast< const char* >( &_e.offset ), sizeof( _e.offset ) );
	_out.write( reinterpret_cast< const char* >( &_e.size ), sizeof( _e.size ) );
	return _out;
}
