#include <fstream>
#include <cassert>
#include <cmath>
#include "Image.h"
#include "BMP.h"

const double PI = 3.14159265;

Image::Image( int _width, int _height, Color _fillColor )
{
	width = _width;
	height = _height;

	rData = new unsigned char[ width * height ];
	gData = new unsigned char[ width * height ];
	bData = new unsigned char[ width * height ];

	Fill( _fillColor );
}

Image::Image( const Image& _other )
{
	assert( this != &_other );

	width = _other.width;
	height = _other.height;

	rData = new unsigned char[ width * height ];
	gData = new unsigned char[ width * height ];
	bData = new unsigned char[ width * height ];

	for( int i = 0; i < width * height; ++i )
	{
		rData[ i ] = _other.rData[ i ];
		gData[ i ] = _other.gData[ i ];
		bData[ i ] = _other.bData[ i ];
	}
}

Image::~Image()
{
	delete[] rData;
	delete[] gData;
	delete[] bData;
}

int Image::GetWidth() const
{
	return width;
}

int Image::GetHeight() const
{
	return height;
}

void Image::Fill( Color _color )
{
	for( int i = 0; i < width * height; ++i )
	{
		rData[ i ] = _color.r;
		gData[ i ] = _color.g;
		bData[ i ] = _color.b;
	}
}

void Image::PutPixel( int _x, int _y, Color _color )
{
	assert( _x >= 0 && _x < width );
	assert( _y >= 0 && _y < height );

	rData[ ( _y * width ) + _x ] = _color.r;
	gData[ ( _y * width ) + _x ] = _color.g;
	bData[ ( _y * width ) + _x ] = _color.b;
}

void Image::DrawRect( int _startX, int _startY, int _endX, int _endY, Color _color )
{
	int firstRow = ( _startY < 0 ) ? 0 : _startY;
	int lastRow = ( _endY > height ) ? height : _endY;

	int firstCol = ( _startX < 0 ) ? 0 : _startX;
	int lastCol = ( _endX > width ) ? width : _endX;

	if( firstCol > width || firstRow > height || lastCol < 0 || lastRow < 0 )
	{
		return;
	}

	for( int row = firstRow; row < lastRow; ++row )
	{
		for( int col = firstCol; col < lastCol; ++col )
		{
			int dataOffset = ( row * width ) + col;
			rData[ dataOffset ] = _color.r;
			gData[ dataOffset ] = _color.g;
			bData[ dataOffset ] = _color.b;
		}
	}
}

void Image::DrawHorizontalLine( int _y, Color _color )
{
	assert( _y >= 0 && _y < height );

	for( int x = 0; x < width; ++x )
	{
		PutPixel( x, _y, _color );
	}
}

void Image::DrawVerticalLine( int _x, Color _color )
{
	assert( _x >= 0 && _x < width );

	for( int y = 0; y < height; ++y )
	{
		PutPixel( _x, y, _color );
	}
}

void Image::DrawCircleOutline( int _posX, int _posY, int _radius, Color _color )
{
	int numCirclePoints = width * 2;
	for( int i = 0; i < numCirclePoints; ++i )
	{
		double angle = 2 * i * PI / numCirclePoints;
		int x = _posX + ( int )( cos( angle ) * _radius );
		int y = _posY + ( int )( sin( angle ) * _radius );

		if( x >= 0 && x < width && y >= 0 && y < height )
		{
			PutPixel( x, y, _color );
		}
	}
}

void Image::DrawCircle( int _posX, int _posY, int _radius, Color _color )
{
	int startX = _posX - _radius;
	int endX = _posX + _radius;
	int startY = _posY - _radius;
	int endY = _posY + _radius;

	startX = std::max( startX, 0 );
	endX = std::min( endX, width - 1 );
	startY = std::max( startY, 0 );
	endY = std::min( endY, height - 1 );

	for( int y = startY; y <= endY; ++y )
	{
		for( int x = startX; x <= endX; ++x )
		{
			int distanceSquared = ( x - _posX )*( x - _posX ) + ( y - _posY )*( y - _posY );
			if( distanceSquared < ( _radius * _radius ) )
			{
				PutPixel( x, y, _color );
			}
		}
	}
}

bool Image::SaveBMP( const std::string& _fileName )
{
	std::ofstream file( _fileName.c_str(), std::ios::binary );
	if( !file )
	{
		return false;
	}

	BMPFileHeader fileHeader( width, height );
	file << fileHeader;

	BMPInfoHeader infoHeader( width, height );
	file << infoHeader;

	int padSize = 4 - ( ( width * 3 ) % 4 );
	if( padSize == 4 ) padSize = 0;

	int lineLength = ( width * 3 ) + padSize;
	unsigned char* lineBuffer = new unsigned char[ lineLength ];
	for( int i = 0; i < lineLength; ++i )
	{
		lineBuffer[ i ] = 0;
	}

	for( int lineIndex = height - 1; lineIndex >= 0; --lineIndex )
	{
		// write line
		for( int i = 0; i < width; ++i )
		{
			lineBuffer[ ( i * 3 )     ] = bData[ ( lineIndex * width ) + i ];
			lineBuffer[ ( i * 3 ) + 1 ] = gData[ ( lineIndex * width ) + i ];
			lineBuffer[ ( i * 3 ) + 2 ] = rData[ ( lineIndex * width ) + i ] ;
		}

		file.write( reinterpret_cast< const char* >( lineBuffer ), lineLength );
	}

	delete[] lineBuffer;

	file.close();
	return true;
}
