#include <algorithm>
#include "Intersection.h"

namespace intersect
{
	bool Circles(
		const Vector2& _posA, Real _radiusA,
		const Vector2& _posB, Real _radiusB )
	{
		Vector2 v = _posB - _posA;
		Real radiiSum = _radiusA + _radiusB;
		return( v.Dot( v ) < radiiSum * radiiSum );
	}

	bool MovingCircles(
		const Vector2& _startA, const Vector2& _endA, Real _radiusA,
		const Vector2& _startB, const Vector2& _endB, Real _radiusB,
		bool* _enter, bool* _exit, Real* _enterTime, Real* _exitTime )
	{
		// fast bounding box test

		if( MovingCirclesBoundingBoxTest( _startA, _endA, _radiusA,
			_startB, _endB, _radiusB ) == false )
		{
			return false;
		}

		// B becomes a sphere with the sum of the radii
		// A becomes a line that moves with velA - velB

		Real radiiSum = _radiusA + _radiusB;

		Vector2 velA = ( _endA - _startA );
		Real lengthA = velA.GetLength();
		Vector2 dirA = velA;
		dirA.Normalize();

		Vector2 velB = ( _endB - _startB );
		Vector2 lineVelocity = velA - velB;
		Real lineLength = lineVelocity.GetLength();
		Vector2 lineDir = lineVelocity;
		lineDir.Normalize();
	
		Vector2 lineStart = _startA;
		Vector2 lineEnd = _startA + lineVelocity;
		Vector2 spherePos = _startB;
		Real sphereRadius = _radiusA + _radiusB;

		bool lineEnters = false;
		bool lineExits = false;
		Vector2 lineEnterPoint;
		Vector2 lineExitPoint;

		if( LineCircle( lineStart, lineEnd, spherePos, sphereRadius,
			&lineEnters, &lineExits, &lineEnterPoint, &lineExitPoint ) == false )
		{
			return false;
		}

		if( _enter )
		{
			*_enter = lineEnters;
		}
		if( _exit )
		{
			*_exit = lineExits;
		}
		if( _enterTime )
		{
			Real enterDistance = ( lineEnterPoint - lineStart ).GetLength();
			*_enterTime = enterDistance / lineLength;
		}
		if( _exitTime )
		{
			Real exitDistance = ( lineExitPoint - lineStart ).GetLength();
			*_exitTime = exitDistance / lineLength;
		}

		return true;
	}


	bool LineCircle(
		const Vector2& _lineStart, const Vector2& _lineEnd,
		const Vector2& _circlePos, Real _circleRadius,
		bool* _enter, bool* _exit, Vector2* _enterPoint, Vector2* _exitPoint )
	{
		if( _enter )
		{
			*_enter = false;
		}

		if( _exit )
		{
			*_exit = false;
		}

		// get line direction and length

		Vector2 lineDir = _lineEnd - _lineStart;
		Real lineLength = lineDir.GetLength();
		lineDir.Normalize();
		
		// project circle center on line

		Vector2 lineStartToCirclePos = _circlePos - _lineStart;
		Real projectionLength = lineDir.Dot( lineStartToCirclePos );
		Vector2 projectedCirclePos = _lineStart + ( lineDir * projectionLength );

		// check circle distance

		Vector2 distanceVector = projectedCirclePos - _circlePos;
		Real distance = distanceVector.GetLength();

		if( distance > _circleRadius )
		{
			return false;
		}

		// get the intersection distances on the line

		Real intersectionLength = ( Real )sqrt( ( _circleRadius * _circleRadius ) - ( distance * distance ) );
		Real enterDistance = projectionLength - intersectionLength;
		Real exitDistance = projectionLength + intersectionLength;

		// check if the intersection points are inside the line

		bool lineEnters = false;
		bool lineExits = false;

		if( enterDistance >= 0 && enterDistance < lineLength )
		{
			lineEnters = true;
		}
		if( exitDistance >= 0 && exitDistance < lineLength )
		{
			lineExits = true;
		}

		if( !lineEnters && !lineExits )
		{
			// the line is inside the circle
			return false;
		}

		if( lineEnters )
		{
			if( _enterPoint )
			{
				*_enterPoint = _lineStart + lineDir * enterDistance;
			}
			if( _enter )
			{
				*_enter = true;
			}
		}
		if( lineExits )
		{
			if( _exitPoint )
			{
				*_exitPoint = _lineStart + lineDir * exitDistance;
			}
			if( _exit )
			{
				*_exit = true;
			}
		}

		return true;
	}
}
