#define NDEBUG

/*********************/

#include <cassert>
#include <algorithm>
#include "Network.h"
#include "Node.h"
#include "Intersection.h"
#include "Bucket.h"

#include <iostream>
#include <fstream>

using namespace std;

#define TWO_SQUARE_ROOT	(1.41421356)

Network::Network( Configuration& _config ) : config( _config ), nodeCheckList( _config.GetNodeCount() )
{
	nodeSpeedMultiplier = config.GetNodeSpeedMultiplier();
	waitingTimeExponent = config.GetWaitingTimeExponent();
	waitingTimeUpperBound = config.GetWaitingTimeUpperBound();
	cellDistanceWeight = config.GetCellDistanceWeight();

	cellLength = ( config.GetNodeRadius() / sqrt( 2.0 ) );
	cellCountPerSide = static_cast< int >( ceil( 1.0 / cellLength ) );
	cellCount = cellCountPerSide * cellCountPerSide;
	
	total_cell_weights = new int[ cellCount ];

	bucketCountPerSide = config.GetBucketCountPerSide();
	bucketCount = bucketCountPerSide * bucketCountPerSide;
	bucketLength = 1.0 / bucketCountPerSide;
	bucketRadius = ( bucketLength * sqrt( 2.0 ) ) / 2.0;

	for( int i = 0; i < bucketCount; ++i )
	{
		buckets.push_back( std::set< Node* >() );
	}
	
	for( int i = 0; i < cellCount; i++ )
		total_cell_weights[ i ] = 0;
}

Network::~Network()
{
	for( size_t i = 0; i < nodes.size(); ++i )
	{
		delete nodes[ i ];
	}
	nodes.clear();
}

size_t Network::AddNode( Real _x, Real _y, int node_count )
{
	size_t newNodeID = nodes.size();

	Node* newNode = new Node( newNodeID, Vector2( _x, _y ), *this, node_count, config.GetNodeRadius() );
	nodes.push_back( newNode );

	return newNodeID;
}

void Network::InitNodes()
{
	for( size_t i = 0; i < nodes.size(); ++i )
	{
		nodes[ i ]->InitCellWeights();
	}
}

size_t Network::GetNodeCount() const
{
	return nodes.size();
}

Node* Network::GetNode( size_t _index )
{
	assert( _index < nodes.size() );
	return nodes[ _index ];
}

Real Network::GetNodeSpeedMultiplier() const
{
	return nodeSpeedMultiplier;
}

Real Network::GetWaitingTimeExponent() const
{
	return waitingTimeExponent;
}

Real Network::GetWaitingTimeUpperBound() const
{
	return waitingTimeUpperBound;
}

int Network::GetCellCount() const
{
	return cellCount;
}

int Network::GetCellCountPerSide() const
{
	return cellCountPerSide;
}

Real Network::GetCellLength() const
{
	return cellLength;
}

void Network::RaiseCellWeight( int index )
{
	total_cell_weights[ index ]++;
}

int Network::GetCellWeight( int index )
{
	return total_cell_weights[ index ];
}

void Network::PrintCellVisits()
{
	ofstream cell_visits( "cell-visits.txt" );

	for( int i = 0; i < cellCountPerSide; i++ )
	{
		for( int j = 0; j < cellCountPerSide; j++ )
		{
			cell_visits << j << ' ' << i << ' ' << total_cell_weights[ j + i*cellCountPerSide ] << endl;
		}
		cell_visits << endl;
	}
}


int Network::GetCellIndexFromPos( const Vector2& _pos ) const
{
	assert( _pos.x >= 0 && _pos.x <= 1 );
	assert( _pos.y >= 0 && _pos.y <= 1 );

	int cellRow = static_cast< int >( _pos.y / cellLength );
	int cellColumn = static_cast< int >( _pos.x / cellLength );

	// cells are indexed in row-major order

	return( cellRow * cellCountPerSide ) + cellColumn;
}

Vector2 Network::GetCellPos( int _cellIndex ) const
{
	assert( _cellIndex >= 0 && _cellIndex < cellCount );

	int cellRow = static_cast< int >( _cellIndex / cellCountPerSide );
	int cellColumn = _cellIndex % cellCountPerSide;

	Real halfCellLength = cellLength / 2.0;
	return Vector2( ( cellColumn * cellLength ) + halfCellLength,
		( cellRow * cellLength ) + halfCellLength );
}

Real Network::GetCellDistanceWeight() const
{
	return config.GetCellDistanceWeight();
}

Vector2 Network::GetRandomPointInCell( int _cellIndex ) const
{
	Vector2 cellCenter = GetCellPos( _cellIndex );

	Real halfCellLength = cellLength / 2.0;

	Real x = random_generator::GetReal( -halfCellLength, halfCellLength );
	Real y = random_generator::GetReal( -halfCellLength, halfCellLength );

	Vector2 point( x + cellCenter.x, y + cellCenter.y );
	if( point.x < 0 )
	{
		point.x = 0;
	}
	else if( point.x >= 1 )
	{
		point.x = 1;
	}
	if( point.y < 0 )
	{
		point.y = 0;
	}
	else if( point.y >= 1 )
	{
		point.y = 1;
	}

	return point;
}

int Network::GetBucketCount() const
{
	return bucketCount;
}

int Network::GetBucketCountPerSide() const
{
	return bucketCountPerSide;
}

Real Network::GetBucketLength() const
{
	return bucketLength;
}

Vector2 Network::GetBucketPos( int _bucketIndex )
{
	assert( _bucketIndex >= 0 && _bucketIndex < bucketCount );

	int bucketRow = static_cast< int >( _bucketIndex / bucketCountPerSide );
	int bucketColumn = _bucketIndex % bucketCountPerSide;

	Real halfBucketLength = bucketLength / 2.0;

	return Vector2( ( bucketColumn * bucketLength ) + halfBucketLength,
		( bucketRow * bucketLength ) + halfBucketLength );
}

void Network::ProcessBuckets( Node* _node, BucketProcessMode _mode )
{
	if( config.GetBucketCountPerSide() == 0 )
	{
		return;
	}

	Real nodeRadius = config.GetNodeRadius();

	Vector2 nodePos = _node->GetPosition();
	Vector2 nodeDest = _node->GetDestination();
	Vector2 nodeVelocity = nodeDest - nodePos;

#ifdef USE_NEW_ALGORITHM_FOR_BUCKETS

	if( _mode == BUCKET_UPDATE )
	{
		for( int i = 0; i < bucketCount; ++i )
		{
			buckets[ i ].erase( _node );
		}
	}
	else
	{
		nodeCheckList.Reset();
	}

	if( ( nodeVelocity.x == 0 ) && ( nodeVelocity.y == 0 ) )
	{
		int firstCol = ( int )( ( nodePos.x - nodeRadius ) / bucketLength );
		int lastCol  = ( int )( ( nodePos.x + nodeRadius ) / bucketLength );

		firstCol = std::max( firstCol, 0 );
		lastCol  = std::min( lastCol, bucketCountPerSide - 1 );

		int firstRow = ( int )( ( nodePos.y - nodeRadius ) / bucketLength );
		int lastRow  = ( int )( ( nodePos.y + nodeRadius ) / bucketLength );

		firstRow = std::max( firstRow, 0 );
		lastRow  = std::min( lastRow, bucketCountPerSide - 1 );

		for( int row = firstRow; row <= lastRow; ++row )
		{
			for( int col = firstCol; col <= lastCol; ++col )
			{
				if( _mode == BUCKET_UPDATE )
				{
					buckets[ ( row * bucketCountPerSide ) + col ].insert( _node );
				}
				else
				{
					nodeCheckList.AddNodes( buckets[ ( row * bucketCountPerSide ) + col ] );
				}
			}
		}

		return;
	}

	if( fabs( nodeVelocity.x ) > fabs( nodeVelocity.y ) )
	{
		// "horizontal" movement: check columns

		int firstCol;
		int lastCol;

		bool movingLeftToRight = ( nodePos.x < nodeDest.x ? true : false );

		if( movingLeftToRight )
		{
			firstCol = ( int )( ( nodePos.x - nodeRadius ) / bucketLength );
			lastCol = 1 + ( int )( ( nodeDest.x + nodeRadius ) / bucketLength );
		}
		else
		{
			firstCol = ( int )( ( nodeDest.x - nodeRadius ) / bucketLength );
			lastCol = 1 + ( int )( ( nodePos.x + nodeRadius ) / bucketLength );
		}

		firstCol = std::max( firstCol, 0 );
		lastCol = std::min( lastCol, bucketCountPerSide );

#ifdef BUCKET_APPROXIMATE_INTERSECTION_POINT

		Real offset = ( nodeRadius * TWO_SQUARE_ROOT );

#else

		Real offset = ( nodeRadius * nodeVelocity.GetLength() ) / fabs( nodeVelocity.x );

#endif

		for( int col = firstCol; col <= lastCol; ++col )
		{
			Real minY, maxY;

			if( col == firstCol )
			{
				Real y = ( movingLeftToRight ? nodePos.y : nodeDest.y );
				minY = y - nodeRadius;
				maxY = y + nodeRadius;
			}
			else if( col == lastCol )
			{
				Real y = ( movingLeftToRight ? nodeDest.y : nodePos.y );
				minY = y - nodeRadius;
				maxY = y + nodeRadius;
			}
			else
			{
				Real colX = ( col * bucketLength );
				Real travelFraction = ( colX - nodePos.x ) / ( nodeDest.x - nodePos.x );
				Real intersectionY = ( travelFraction * nodeVelocity.y ) + nodePos.y;

				minY = intersectionY - offset;
				maxY = intersectionY + offset;
			}

			int firstBucketRow = ( int )( minY / bucketLength );
			int lastBucketRow = ( int )( maxY / bucketLength );

			firstBucketRow = std::max( firstBucketRow, 0 );
			lastBucketRow = std::min( lastBucketRow, bucketCountPerSide - 1 );

			for( int row = firstBucketRow; row <= lastBucketRow; ++row )
			{
				if( _mode == BUCKET_UPDATE )
				{
					if( col > firstCol )
					{
						buckets[ ( row * bucketCountPerSide ) + col - 1 ].insert( _node );
					} 
					if( col < lastCol )
					{
						buckets[ ( row * bucketCountPerSide ) + col ].insert( _node );
					}
				}
				else
				{
					if( col > firstCol )
					{
						nodeCheckList.AddNodes( buckets[ ( row * bucketCountPerSide ) + col - 1 ] );
					}
					if( col < lastCol )
					{
						nodeCheckList.AddNodes( buckets[ ( row * bucketCountPerSide ) + col ] );
					}
				}
			}
		}
	}
	else
	{
		// "vertical" movement: check rows

		int firstRow;
		int lastRow;

		bool descending = ( nodePos.y < nodeDest.y ? true : false );

		if( descending )
		{
			firstRow = ( int )( ( nodePos.y - nodeRadius )/ bucketLength );
			lastRow = 1 + ( int )( ( nodeDest.y + nodeRadius ) / bucketLength );
		}
		else
		{
			firstRow = ( int )( ( nodeDest.y - nodeRadius )/ bucketLength );
			lastRow = 1 + ( int )( ( nodePos.y + nodeRadius ) / bucketLength );
		}

		firstRow = std::max( firstRow, 0 );
		lastRow = std::min( lastRow, bucketCountPerSide );

#ifdef BUCKET_APPROXIMATE_INTERSECTION_POINT

		Real offset = ( nodeRadius * TWO_SQUARE_ROOT );

#else

		Real offset = ( nodeRadius * nodeVelocity.GetLength() ) / fabs( nodeVelocity.y );

#endif

		for( int row = firstRow; row <= lastRow; ++row )
		{
			Real minX, maxX;

			if( row == firstRow )
			{
				Real x = ( descending ? nodePos.x : nodeDest.x );
				minX = x - nodeRadius;
				maxX = x + nodeRadius;
			}
			else if( row == lastRow )
			{
				Real x = ( descending ? nodeDest.x : nodePos.x );
				minX = x - nodeRadius;
				maxX = x + nodeRadius;
			}
			else
			{
				Real rowY = row * bucketLength;
				Real travelFraction = ( rowY - nodePos.y ) / ( nodeDest.y - nodePos.y );
				Real intersectionX = ( travelFraction * nodeVelocity.x ) + nodePos.x;

				minX = intersectionX - offset;
				maxX = intersectionX + offset;
			}

			int firstBucketCol = ( int )( minX / bucketLength );
			int lastBucketCol = ( int )( maxX / bucketLength );

			firstBucketCol = std::max( firstBucketCol, 0 );
			lastBucketCol = std::min( lastBucketCol, bucketCountPerSide - 1 );

			for( int col = firstBucketCol; col <= lastBucketCol; ++col )
			{
				if( _mode == BUCKET_UPDATE )
				{
					if( row > firstRow )
					{
						buckets[ ( ( row - 1 ) * bucketCountPerSide ) + col ].insert( _node );
					}
					if( row < lastRow )
					{
						buckets[ ( row * bucketCountPerSide ) + col ].insert( _node );
					}
				}
				else
				{
					if( row > firstRow )
					{
						nodeCheckList.AddNodes( buckets[ ( ( row - 1 ) * bucketCountPerSide ) + col ] );
					}
					if( row < lastRow )
					{
						nodeCheckList.AddNodes( buckets[ ( row * bucketCountPerSide ) + col ] );
					}
				}
			}
		}
	}

#else

	if( _mode == BUCKET_UPDATE )
	{
		for( int i = 0; i < bucketCount; ++i )
		{
			if( CheckPassage( i, _node ) )
			{
				buckets[ i ].insert( _node );
			}
			else //if( _node->GetState() != Node::MOVING )
			{
				buckets[ i ].erase( _node );
			}
		}
	}
	else
	{
		nodeCheckList.Reset();

		for( int i = 0; i < bucketCount; ++i )
		{
			if( CheckPassage( i, _node ) )
			{
				nodeCheckList.AddNodes( buckets[ i ] );
			}
		}
	}

#endif
}


bool Network::CheckPassage( int _bucketIndex, Node* _node )
{
	Vector2 bucketPos = GetBucketPos( _bucketIndex );
	Vector2 nodePos = _node->GetPosition();
	Vector2 nodeDest = _node->GetDestination();
	Real nodeRadius = config.GetNodeRadius();
	Real bucketHalfLength = bucketLength / 2.0;

#ifdef USE_BOUNDING_BOX_BUCKETS

	if( intersect::BoundingBoxes(
		Vector2( bucketPos.x - bucketHalfLength, bucketPos.y - bucketHalfLength ),
		Vector2( bucketPos.x + bucketHalfLength, bucketPos.y + bucketHalfLength ),
		Vector2( std::min( nodePos.x, nodeDest.x ) - nodeRadius, std::min( nodePos.y, nodeDest.y ) - nodeRadius ),
		Vector2( std::max( nodePos.x, nodeDest.x ) + nodeRadius, std::max( nodePos.y, nodeDest.y ) + nodeRadius ) ) )
	{
		return true;
	}

	return false;

#else

	if( intersect::Circles( nodePos, nodeRadius, GetBucketPos( _bucketIndex ), bucketRadius ) )
	{
		return true;
	}
	else
	{
		if( _node->GetState() == Node::MOVING )
		{
			if( intersect::MovingCircles(
				nodePos, nodeDest, nodeRadius,
				bucketPos, bucketPos, bucketRadius, NULL, NULL, NULL, NULL ) )
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}

#endif

}

NodeCheckList* Network::GetNodesToCheck( Node* _node )
{
	ProcessBuckets( _node, BUCKET_CREATE_NODE_CHECK_LIST );

#ifdef BUCKET_SORT_NODELISTS

	nodeCheckList.Sort();

#endif

	return &nodeCheckList;
}


std::set< Node* >& Network::TEMP_GetBucket( int i )
{
	return buckets[ i ];
}

int Network::GetNumberOfNodes()
{
	return config.GetNodeCount();
}


