#define NDEBUG

/*********************/

#include <cassert>
#include <iostream>
#include <cstdio>
#include "Node.h"
#include "Bucket.h"
#include "Intersection.h"
#include "math.h"

using namespace std;

Node::Node( size_t _id, Vector2 _homePos, Network& _network, int node_count, Real _nodeRadius  ) :
	network( _network ),
	cellDistanceWeight( _network.GetCellDistanceWeight() ),
	cellWeights( _network.GetCellCount() ), neighbours_list( node_count ),
	number_of_visits( _network.GetCellCount() ), number_of_nodes_seen( _network.GetCellCount() ), 
	number_of_nodes_seen_last_visit( _network.GetCellCount() )
{

	state		= Node::NEW;
	id		= _id;
	pos		= _homePos;
	posTime		= 0;
	dest		= _homePos;
	speed		= 0;
	waitTime	= 0;

	home		= _homePos;

	currentCell	= network.GetCellIndexFromPos( _homePos );
	destinationCell	= currentCell;
	contactCount	= 0;
	nodeRadius	= _nodeRadius;
	numberOfNodes	= network.GetNumberOfNodes();
	densita		= M_PI * nodeRadius * nodeRadius * numberOfNodes;
	


	for( size_t i = 0; i < cellWeights.size(); ++i )
	{
		cellWeights[ i ]			= 0.0;
		number_of_visits[ i ]			= 0;
		number_of_nodes_seen[ i ]		= 0;
		number_of_nodes_seen_last_visit[ i ]	= 0;
		
		if( (cellDistanceWeight == 0.0) )
		{
			number_of_nodes_seen[ i ]	= 1;
			number_of_visits[ i ]		= 1;
			number_of_nodes_seen_last_visit[ i ]	= 1;
		}
	}
	
	
	for( size_t i = 0; i < neighbours_list.size(); ++i )
	{
		neighbours_list[ i ] = 0;
	}
}

void Node::UpdatePosition( Vector2 _pos, Real _time )
{
	pos = _pos;
	posTime = _time;
}

void Node::MoveToRandomDestination()
{
	SetCellWeight( currentCell, number_of_nodes_seen_last_visit[ currentCell ] );
	number_of_nodes_seen_last_visit[ currentCell ] = 0;

	destinationCell = ChooseDestinationCell();
	assert( destinationCell != currentCell );
	
	network.RaiseCellWeight( destinationCell );
	
	Vector2 destinationPoint = network.GetRandomPointInCell( destinationCell );

	state		= Node::MOVING;
	dest		= destinationPoint;
	speed		= ( destinationPoint - pos ).GetLength() * network.GetNodeSpeedMultiplier();
	waitTime	= 0;

	network.ProcessBuckets( this, BUCKET_UPDATE );
}

void Node::WaitRandomTime()
{
	state		= Node::WAITING;
	dest		= pos;
	speed		= 0;
	waitTime	= ComputeRandomWaitingTime();

	currentCell = destinationCell;
	number_of_visits[ currentCell ]++;

	network.ProcessBuckets( this, BUCKET_UPDATE  );
}

void Node::Meet( Node* _other, Real currentTime, Real knowingTime )
{
	if( ( state == Node::WAITING && _other->GetState() == Node::MOVING ) )
	{
		if( intersect::Circles( pos, nodeRadius, _other->GetDestination(), nodeRadius ) )
		{
			number_of_nodes_seen_last_visit[ currentCell ]++;
			if( currentTime > knowingTime )
				neighbours_list[ _other->GetID() ]++;
		}
	}
	else
	{
		if( ( state == Node::MOVING && _other->GetState() == Node::MOVING ) )
		{
			if( intersect::Circles( dest, nodeRadius, _other->GetDestination(), nodeRadius ) )
			{
				number_of_nodes_seen_last_visit[ destinationCell ]++;
				if( currentTime > knowingTime )
					neighbours_list[ _other->GetID() ]++;
			}
		}
		
		if( ( state == Node::MOVING && _other->GetState() == Node::WAITING ) )
		{
			if( intersect::Circles( dest, nodeRadius, _other->GetPosition(), nodeRadius ) )
			{
				number_of_nodes_seen_last_visit[ destinationCell ]++;
				if( currentTime > knowingTime )
					neighbours_list[ _other->GetID() ]++;
			}
		}
	}
}

int Node::GetNeighbourContactTime(int index)
{
	return neighbours_list[ index ];
}

void Node::Leave( Node* _other )
{
	return;
}

size_t Node::GetID() const
{
	return id;
}

Node::State Node::GetState() const
{
	return state;
}

Vector2 Node::GetPosition() const
{
	return pos;
}

Vector2 Node::GetDestination() const
{
	return dest;
}

Vector2 Node::GetHomePosition() const
{
	return home;
}

//-----------------------------------------------------------------------------
//
//	Name:	Node::GetPositionAbsoluteTime()
//	
//	Desc:	Return the absolute time when the node was in the current position.
//
//-----------------------------------------------------------------------------
Real Node::GetPositionAbsoluteTime() const
{
	return posTime;
}

//-----------------------------------------------------------------------------
//
//	Name:	Node::GetTravelTime()
//	
//	Desc:	Return the travel time of the node. This is the time to reach the 
//			destination if the node is moving, or the waiting time if the node
//			is waiting.
//
//-----------------------------------------------------------------------------
Real Node::GetTravelTime() const
{
	if( state == Node::WAITING )
	{
		return waitTime;
	}
	else if( state == Node::MOVING )
	{
		Real moveDistance = ( dest - pos ).GetLength();
		return( moveDistance / speed );
	}
	else
	{
		return 0;
	}
}

//-----------------------------------------------------------------------------
//
//	Name:	Node::ComputePositionAtTime()
//	
//	Desc:	Return the node position at the given absolute time.
//
//-----------------------------------------------------------------------------
Vector2 Node::ComputePositionAtTime( Real _absTime ) const
{
	Vector2 dir = ( dest - pos );
	dir.Normalize();

	return( pos + ( dir * speed * ( _absTime - posTime ) ) );
}

//-----------------------------------------------------------------------------
//
//	Cell weights
//
//-----------------------------------------------------------------------------
void Node::InitCellWeights()
{
	for( size_t i = 0; i < cellWeights.size(); ++i )
	{
		SetCellWeight( ( int )i, 0 );
	}
}


void Node::SetCellWeight( int _cellIndex, int _seen )
{
	number_of_nodes_seen[ _cellIndex ] += _seen;
	
	Real distanceValue = DistanceFunction( _cellIndex );
	Real seenValue = SeenFunction( _cellIndex );

	cellWeights[ _cellIndex ] = ( cellDistanceWeight * distanceValue ) +
		( ( 1.0 - cellDistanceWeight ) * seenValue );
}


Real Node::GetCellWeight( int _cellIndex )
{
	return cellWeights[ _cellIndex ];
}


bool Node::operator<( const Node& _other )
{
	return( id < _other.id );
}

//-----------------------------------------------------------------------------
//
//	Name:	Node::ComputeRandomWaitingTime()
//	
//	Desc:	Compute a random waiting time with an upper bounded power law
//			distribution.
//
//			The distribution function is f(x) = (slope-1) * x^(-slope).
//			The integral from 1 to +inf of f(x) is equal to 1.
//
//			A random number y is uniformly chosen in [0,1]. This number
//			represents the area under f(x) starting from x=1.
//			This area is the integral from 1 to z of f(x) and is equal to
//			1 - z^( -slope+1 ). So:
//
//			y = 1 - z^( -slope+1 )
//
//			Thus:
//
//			z = ( 1-y )^( 1 / ( -slope+1 ) )
//
//			z is distributed in [1,+inf) as the power law f(x).
//
//-----------------------------------------------------------------------------
Real Node::ComputeRandomWaitingTime() const
{
	Real slope = network.GetWaitingTimeExponent();
	Real upperBound = network.GetWaitingTimeUpperBound();

	// choose a random number uniformly in [0,1]

	Real y = random_generator::GetReal();

	// convert to a random number in [1,+inf) with power law distribution

	Real exponent = 1.0 / ( -slope + 1 );
	Real time = pow( ( double )( 1.0 - y ), ( double )exponent );

	// apply an upper bound

	if( time > upperBound )
	{
		time = upperBound;
	}

	return time;
}

//-----------------------------------------------------------------------------
//
//	Name:	Node::DistanceFunction()
//	
//	Desc:	Compute the distance() function of SWIM
//
//-----------------------------------------------------------------------------
Real Node::DistanceFunction( int _cellIndex ) const
{

// dist(Ci) = (1 / (1 + k*D(i))^2 ) / max_j{ 1 / (1 + k*D(j))^2 }

	Real k = 1 / nodeRadius, max = 0.0;
	Real distance = ( home - network.GetCellPos( _cellIndex ) ).GetLength();
	Real denominator = 1 + ( k * distance );
	denominator *= denominator;
	Real inverse_distance_Ci = 1 / denominator;
	
	for( int j = 0; j < cellWeights.size(); j++ )
	{
		Real distance_Cj = ( home - network.GetCellPos( j ) ).GetLength();
		
		Real denominator_Cj = 1 + ( k * distance_Cj );
		denominator_Cj *= denominator_Cj;
		Real inverse_distance_Cj = 1 / denominator_Cj;
		
		if( inverse_distance_Cj > max )
			max = inverse_distance_Cj;
	}
	
	return inverse_distance_Ci / max;

}

//-----------------------------------------------------------------------------
//
//	Name:	Node::SeenFunction()
//	
//	Desc:	Compute the seen() function of SWIM
//
//-----------------------------------------------------------------------------
Real Node::SeenFunction( int _cellIndex ) const
{
	assert( _cellIndex >= 0 );
	
	// Versione 5: seen = ( 1 + 1/d * visti_Ci ) / max_j{ 1 + 1/d * visti_Cj }
	
	int visti_Ci = number_of_nodes_seen[ _cellIndex ];
	
	if( visti_Ci == 0 )
		return 1.0 / cellWeights.size();
	
	Real nominator = 1 + visti_Ci / densita;
	
	Real max = 0.0;
	for( int j = 0; j < cellWeights.size(); j++ )
	{
		int visti_Cj = number_of_nodes_seen[ j ];
		Real nominator_j = 1 + visti_Cj / densita;
		
		if( nominator_j > max )
			max = nominator_j;
	}
		

	return ( nominator / max );
	
}

//-----------------------------------------------------------------------------
//
//	Name:	Node::ChooseDestinationCell()
//
//	Desc:	Choose a random cell proportionally with the cell weights.
//			Assumes that the sum of the weights is 1.
//
//-----------------------------------------------------------------------------
int Node::ChooseDestinationCell() const
{
	int cell = 0;

	Real weightSum = 0;
	for( size_t i  = 0; i < cellWeights.size(); ++i )
	{
		if( i != currentCell )
		{
			weightSum += cellWeights[ i ];
		}
	}
	
	Real randomReal = random_generator::GetReal( 0, weightSum );

	weightSum = 0;
	for( size_t i = 0; i < cellWeights.size(); ++i )
	{
		if( i != currentCell )
		{
			Real prevWeightSum = weightSum;
			weightSum += cellWeights[ i ];
			if( randomReal >= prevWeightSum && randomReal <= weightSum )
			{
				cell = static_cast< int >( i );
			}
		}
	}

	return cell;
}


int Node::GetDestinationCell()
{
	return destinationCell;
}


void Node::PrintCellWeights()
{
	for( int i = 0; i < cellWeights.size(); i++ )
	{
		printf( "%lf\t%lf\t%lf\n", SeenFunction( i ), DistanceFunction( i ), cellWeights[i] );
//		printf( "%lf\n", cellWeights[i] );
//		printf( "%lf\n", DistanceFunction( i ) );
	}
}

