#define NDEBUG

/*********************/

#include <cassert>
#include <iostream>
#include <string>
#include <sstream>
#include <ctime>
#include <fstream>
#include "Simulation.h"
#include "Random.h"
#include "Intersection.h"
#include "Bucket.h"

#include <iostream>
#include <iomanip>
#include <ios>
using namespace std;

struct ContactInfo
{
	ContactInfo() : checkCount( 0 ), eventCount( 0 ) {}
	int checkCount;
	int eventCount;
};
static ContactInfo* contactInfo;

Simulation::Simulation( Network& _n, const Configuration& _config ) : network( _n ), config( _config )
{
	currentTime		= 0;
	
	knowingTime		= config.GetKnowingTime();
	maxTime			= config.GetSimulationTime();
	meetRadius		= config.GetNodeRadius();	
	
	meetInPlace = new bool*[config.GetNodeCount()];
	for( size_t i = 0; i < config.GetNodeCount(); ++i )
		meetInPlace[i] = new bool[config.GetNodeCount()];
	
	tracefile.open( config.GetTraceFileName().c_str(), ios::trunc );
	logfile.open( "log", ios::trunc );
}

Simulation::~Simulation()
{
	for( size_t i = 0; i < network.GetNodeCount(); ++i )
		delete[] meetInPlace[i];
	delete[] meetInPlace;
	
	tracefile.close();
	logfile.close();
}

void Simulation::Start()
{
	
	for( size_t i = 0; i < network.GetNodeCount(); ++i )
		for( size_t j = 0; j < network.GetNodeCount(); ++j )
		{
			meetInPlace[i][j] = false;
			meetInPlace[j][i] = false;
		}
	
	logfile << "-------- print initial position of nodes... --------" << endl;
	/* Print the initial position of nodes */
	ofstream initialPositionFile("initial-positions");
	
	for( size_t i = 0; i < network.GetNodeCount(); ++i )
	{
		Node* node = network.GetNode( i );
		
		initialPositionFile << node->GetID() << " " << node->GetHomePosition().x 
		<< " " << node->GetHomePosition().y << endl;
	}
	
	initialPositionFile.close();
	
	logfile << "-------- check for initial contacts... --------" << endl;	
	// check for initial contacts
	
	for( size_t i = 0; i < network.GetNodeCount(); ++i )
	{
		logfile << "              check initial contacts of node " << i << endl;
		Node* node = network.GetNode( i );
		for( size_t j = i + 1; j < network.GetNodeCount(); ++j )
		{
			Node* otherNode = network.GetNode( j );
			if( intersect::Circles( node->GetPosition(), meetRadius,
								   otherNode->GetPosition(), meetRadius ) )
			{
				eventQueue.push( Event( Event::MEET, node, otherNode, 0 ) );
			}
		}
	}
	
	logfile << "------- create initial events... --------" << endl;	
	// create initial events
	
	for( size_t i = 0; i < network.GetNodeCount(); ++i )
	{
		Node* node = network.GetNode( i );
		
		eventQueue.push( Event( Event::START_WAITING, node, NULL, 0 ) );
	}
	
	// print random seed
	
	if( config.IsShowProgressEnabled() )
	{
#ifndef BUCKET_TEST
		std::cerr << "- seed: " << random_generator::GetSeed() << std::endl;
#endif
	}
	
	// write trace header
	
	/*
	 if( config.IsShowTraceEnabled() )
	 {
	 std::cout << "NETWORK_NODE_NO " << network.GetNodeCount() << std::endl;
	 std::cout << "NETWORK_BIRTH_TIME " << 0 << std::endl;
	 std::cout << "NETWORK_DEATH_TIME " << maxTime << std::endl;
	 }
	 if( config.IsSaveTraceFileEnabled() && ( traceFile != NULL ) )
	 {
	 *traceFile << "NETWORK_NODE_NO " << network.GetNodeCount() << std::endl;
	 *traceFile << "NETWORK_BIRTH_TIME " << 0 << std::endl;
	 *traceFile << "NETWORK_DEATH_TIME " << maxTime << std::endl;
	 }
	 */
	
	// perform simulation
	
	contactInfo = new ContactInfo[ network.GetNodeCount() ];
	
	clock_t simulationStartTicks = clock();
	
	while( true )
	{
		if( eventQueue.empty() )
		{
			break;
		}
		
		Event e = eventQueue.top();
		eventQueue.pop();
		currentTime = e.time;
		
		if( currentTime >= knowingTime + maxTime )
		{
			break;
		}
		
		// print progress info
		
		if( config.IsShowProgressEnabled() )
		{
			static Real lastPrintTime = 0;
			static int progressPercentage = 0;
			static int progressCounter = 0;
			Real elapsedTime = 0;
			
			int newProgressPercentage = ( int )( ( currentTime / ( maxTime + knowingTime ) ) * 100.0 );
			
			if( newProgressPercentage > progressPercentage )
			{
				progressCounter += newProgressPercentage - progressPercentage;
				progressPercentage = newProgressPercentage;
				
				if( progressCounter >= config.GetProgressStep() )
				{
					elapsedTime = ( clock() - simulationStartTicks ) / ( Real )CLOCKS_PER_SEC;
					std::cerr
					<< "- progress: " << progressPercentage << "%"
					<< " - elapsed time: " << elapsedTime
					<< " seconds" << std::endl;
					
					logfile << "- progress: " << progressPercentage << "%"
					<< " - elapsed time: " << elapsedTime
					<< " seconds" << endl;
					
					progressCounter = 0;
				}
			}
		}
		
		// print event
		
		/* MP = Meet in Place is the meeting between two nodes both in their destinations
		 MM = Meet in Movement is the meeting between two nodes that at least one is in moving state
		 
		 Respectively LP and LM stands for Leave in Place and Leave in Movement			
		 */
		
		if( currentTime > knowingTime)
		{
			if( e.type == Event::MEET )
			{
				if( ( e.firstNode->GetState() == Node::WAITING && e.secondNode->GetState() == Node::MOVING ) )
				{
					if( intersect::Circles( e.firstNode->GetPosition(), meetRadius, e.secondNode->GetDestination(), meetRadius ) )
					{
						meetInPlace[e.firstNode->GetID()][e.secondNode->GetID()] = true;
						meetInPlace[e.secondNode->GetID()][e.firstNode->GetID()] = true;
						PrintEvent( e, "MP" );
					}
					else
						PrintEvent( e, "MM" );
				}
				else
				{
					if( (e.firstNode->GetState() == Node::MOVING && e.secondNode->GetState() == Node::MOVING ) )
					{
						if( intersect::Circles( e.firstNode->GetDestination(), meetRadius, e.secondNode->GetDestination(), meetRadius ) )
						{
							// Find the time in which at least one node reaches the desinations
							Real minTravelTime = min( e.firstNode->GetPositionAbsoluteTime() + e.firstNode->GetTravelTime(), 
													 e.secondNode->GetPositionAbsoluteTime() + e.secondNode->GetTravelTime() );
							
							// If in that time the nodes are still seeing each other we print MP, otherwise we print MM
							if( intersect::Circles( e.firstNode->ComputePositionAtTime(minTravelTime), meetRadius, 
												   e.secondNode->ComputePositionAtTime(minTravelTime), meetRadius ) )
							{
								meetInPlace[e.firstNode->GetID()][e.secondNode->GetID()] = true;
								meetInPlace[e.secondNode->GetID()][e.firstNode->GetID()] = true;
								PrintEvent( e, "MP" );
							}
							else
								PrintEvent( e, "MM" );
						}
						else
							PrintEvent( e, "MM" );
					}
					
					else if( ( e.firstNode->GetState() == Node::MOVING && e.secondNode->GetState() == Node::WAITING ) )
					{
						if( intersect::Circles( e.firstNode->GetDestination(), meetRadius, e.secondNode->GetPosition(), meetRadius ) )
						{
							meetInPlace[e.firstNode->GetID()][e.secondNode->GetID()] = true;
							meetInPlace[e.secondNode->GetID()][e.firstNode->GetID()] = true;
							PrintEvent( e, "MP" );
						}
						else
							PrintEvent( e, "MM" );
					}
				}
			}
			else if ( e.type == Event::LEAVE )
			{
				if( meetInPlace[e.firstNode->GetID()][e.secondNode->GetID()]
						|| meetInPlace[e.secondNode->GetID()][e.firstNode->GetID()] )
				{
					meetInPlace[e.firstNode->GetID()][e.secondNode->GetID()] = false;
					meetInPlace[e.secondNode->GetID()][e.firstNode->GetID()] = false;
					PrintEvent( e, "LP" );
				}
				else
					PrintEvent( e, "LM" );

			}
			else if ( config.IsShowOnlyMeetAndLeaveEnabled() == false )
				PrintEvent(e);
		}
		
		// handle event
		
		switch( e.type )
		{
			case Event::START_MOVING:
			{
				e.firstNode->UpdatePosition( e.firstNode->GetDestination(), currentTime );
				e.firstNode->MoveToRandomDestination();
				eventQueue.push( Event( Event::END_MOVING, e.firstNode, NULL, e.time + e.firstNode->GetTravelTime() ) );
				
				CheckContacts( e.firstNode );
				break;
			}
				
			case Event::START_WAITING:
			{
				e.firstNode->UpdatePosition( e.firstNode->GetDestination(), currentTime );
				e.firstNode->WaitRandomTime();
				Real timeToWait = e.firstNode->GetTravelTime();
				eventQueue.push( Event( Event::END_WAITING, e.firstNode, NULL, e.time + timeToWait ) );
				
				CheckContacts( e.firstNode );
				break;
			}
				
			case Event::END_MOVING:
			{
				eventQueue.push( Event( Event::START_WAITING, e.firstNode, NULL, e.time ) );
				break;
			}
				
			case Event::END_WAITING:
			{
				eventQueue.push( Event( Event::START_MOVING, e.firstNode, NULL, e.time ) );
				break;
			}
				
			case Event::MEET:
			{
				e.firstNode->Meet( e.secondNode, currentTime, knowingTime );
				e.secondNode->Meet( e.firstNode, currentTime, knowingTime );
				break;
			}
				
			case Event::LEAVE:
			{
				e.firstNode->Leave( e.secondNode );
				e.secondNode->Leave( e.firstNode );
				break;
			}
				
		}
	}
	
	// Print the weights that a given node has calculated for all the cells
	
	//	Node* temp_node = network.GetNode( 0 );
	//	temp_node->PrintCellWeights();
	
	
	// Print total number of visits per cell counted during simulation
	
	//	network.PrintCellVisits();
	
	if( config.IsShowProgressEnabled() )
	{
#ifndef BUCKET_TEST
		Real simulationTime = ( ( clock() - simulationStartTicks ) / ( Real )CLOCKS_PER_SEC );
		std::cerr << "- simulation completed - total elapsed time: "
		<< simulationTime << " seconds" << std::endl;
		
		logfile << "- simulation completed - total elapsed time: "
		<< simulationTime << " seconds" << std::endl;
#else
		std::cerr << ( ( clock() - simulationStartTicks ) / ( Real )CLOCKS_PER_SEC ) << std::endl;
#endif
		
		Real average = 0;
		for( size_t i = 0; i < network.GetNodeCount(); ++i )
		{
			average += contactInfo[ i ].checkCount / ( Real )contactInfo[ i ].eventCount;
		}
		average /= network.GetNodeCount();
		
#ifndef BUCKET_TEST
		std::cerr << "average checks per node: " << average << std::endl;
#else
		std::cerr << average << std::endl;
#endif
	}
	
}

//-----------------------------------------------------------------------------
//
//	Simulation::CheckContacts()
//
//	Check if the node will meet other nodes in its current movement.
//
//-----------------------------------------------------------------------------
void Simulation::CheckContacts( Node* node )
{
	++contactInfo[ node->GetID() ].eventCount;
	
	if( config.GetBucketCountPerSide() > 0 )
	{
		// buckets
		
		NodeCheckList* list = network.GetNodesToCheck( node );
		for( size_t i = 0; i < list->GetNodeCount(); ++i )
		{
			CheckContactWithNode( node, list->GetNode( i ) );
			++contactInfo[ node->GetID() ].checkCount;
		}
	}
	else
	{
		// no buckets, check with every other node
		
		for( size_t i = 0; i < network.GetNodeCount(); ++i )
		{
			Node* otherNode = network.GetNode( i );
			if( node != otherNode )
			{
				CheckContactWithNode( node, otherNode );
				++contactInfo[ node->GetID() ].checkCount;
			}
		}
	}
}

//-----------------------------------------------------------------------------
//
//	Simulation::CheckContactWithNode()
//
//	Check for contact between two nodes. Return true if the nodes will meet,
//	else false. Also add the MEET and LEAVE events in the queue.
//
//-----------------------------------------------------------------------------
bool Simulation::CheckContactWithNode( Node* node, Node* otherNode )
{
	// compute the node start and end times
	
	Real timeStartA = node->GetPositionAbsoluteTime();
	Real timeEndA = timeStartA + node->GetTravelTime();
	Real timeStartB = otherNode->GetPositionAbsoluteTime();
	Real timeEndB = timeStartB + otherNode->GetTravelTime();
	
	// compute the common time area
	
	Real timeStart = ( timeStartA >= timeStartB ? timeStartA : timeStartB );
	Real timeEnd = ( timeEndA <= timeEndB ? timeEndA : timeEndB );
	
	assert( ( timeEnd - timeStart ) >= 0 );
	
	// compute the node start and end positions
	
	Vector2 startA = node->ComputePositionAtTime( timeStart );
	Vector2 endA = node->ComputePositionAtTime( timeEnd );
	Vector2 startB = otherNode->ComputePositionAtTime( timeStart );
	Vector2 endB = otherNode->ComputePositionAtTime( timeEnd );
	
	// check and handle contacts
	
	Real meetFraction = 0;
	Real leaveFraction = 0;
	
	bool meet = false;
	bool leave = false;
	
	if( intersect::MovingCircles(
								 startA, endA, meetRadius,
								 startB, endB, meetRadius,
								 &meet, &leave, &meetFraction, &leaveFraction ) )
	{
		if( meet )
		{
			Real meetTime = meetFraction * ( timeEnd - timeStart );
			eventQueue.push( Event( Event::MEET, node, otherNode, currentTime + meetTime ) );
		}
		if( leave )
		{
			Real leaveTime = leaveFraction * ( timeEnd - timeStart );
			eventQueue.push( Event( Event::LEAVE, node, otherNode, currentTime + leaveTime ) );
		}
	}
	
	return meet;
}


void Simulation::PrintEvent( Event e, const char* eventType )
{
	tracefile << std::fixed << std::setprecision(3)
	<< e.getTime() << ' ' << eventType << ' '
	<< std::setw(4)	<< e.firstNode->GetID() << ' ' 
	<< std::setw(4) << e.secondNode->GetID() << ' '
	<< (e.firstNode->ComputePositionAtTime(currentTime)).GetX() << ' '
	<< (e.firstNode->ComputePositionAtTime(currentTime)).GetY() << ' '
	<< (e.secondNode->ComputePositionAtTime(currentTime)).GetX() << ' '
	<< (e.secondNode->ComputePositionAtTime(currentTime)).GetY()
	<< endl;
}

void Simulation::PrintEvent( Event e )
{
	const char* eventType;
	switch( e.type )
		{
			case Event::START_MOVING:
			{
				eventType = "SM";
				break;
			}
				
			case Event::START_WAITING:
			{
				eventType = "SW";
				break;
			}
				
			case Event::END_MOVING:
			{
				eventType = "EM";	
				break;
			}
				
			case Event::END_WAITING:
			{
				eventType = "EW";
				break;
			}
		}

	tracefile << std::fixed << std::setprecision(3)
	<< e.getTime() << ' ' << eventType << ' '
	<< std::setw(4)	<< e.firstNode->GetID() << ' ' 
	<< (e.firstNode->ComputePositionAtTime(currentTime)).GetX() << ' '
	<< (e.firstNode->ComputePositionAtTime(currentTime)).GetY() << ' '
	<< endl;
}

Real Simulation::min( Real a, Real b )
{
	if( a < b ) return a;
	else return b;
}




